/* -*- C -*- (for emacs to use C-mode)
 * bib2html.lex
 * This file implements the flex parser for bibfiles
 * and the C code to convert that into an html file.
 * Author: Kiri Wagstaff, wkiri@cs.cornell.edu
 */

%{
  #include <string.h>
  #include <assert.h>
  #include <stdio.h>
  #include <unistd.h> /* for getlogin() */
  #define YYSTYPE char* /* this doesn't make it into bib2html.h on its own */
  #define BUFSIZE 512
  #include "bib2html.h" /* interface to yacc */
  extern int yydebug;
  void clean(char* str);
  char* find_string(char* str);
  FILE* fout; 
  int linenumber = 1;
%}
    int curCat = 0, i;
    char c;

WS   [ \t]

%option noyywrap
%%
{WS}+ { yylval = strdup(yytext); return TOK_WS; }
\n { linenumber++; yylval = strdup(" ");  return TOK_WS; }
"%"[^%][^\n]* { yylval = strdup(yytext); return TOK_COMMENT; }
^"@"[sS]"tring{"[^}]*"}"  /* consume strings */
^"@"[^{(]+"{"[^,}]+ {
  *(strstr(yytext, "{")) = 0;
  yylval = strdup(yytext+1);
  if (yydebug) printf("\n B: Got type <%s>\n", yylval);
  return TOK_OPENCURLY;
}
^"@"[^({]+"("[^,)]+ {
  *(strstr(yytext, "(")) = 0;
  yylval = strdup(yytext+1);
  if (yydebug) printf("\n B: Got type <%s>\n", yylval);
  return TOK_OPAREN;
}
"{" { return TOK_OPENCURLY; }
"}" { return TOK_CLOSECURLY; }
")" { return TOK_CPAREN; }
\"  { yylval = strdup(yytext); return TOK_QUOTE; }
\\  { yylval = strdup(yytext); return TOK_SLASH; }
,   { yylval = strdup(yytext); return TOK_COMMA; }
=   { yylval = strdup(yytext); return TOK_EQ; }
~   { yylval = strdup(yytext); return TOK_TILDE; }
^"%%% "[^%]+"%%%"{WS}*\n { /* Output section headings */
  char* ptr;
  linenumber++; /* must terminate with newline */
  ptr = rindex(yytext, '%'); ptr++; *ptr = 0;
  yyleng = strlen(yytext);
  if (yytext[yyleng-4] == ' ') yytext[yyleng-4] = 0;
  else yytext[yyleng-3] = 0;
  yylval = (char*)malloc(BUFSIZE); assert(yylval);
  sprintf(yylval, "<A NAME=topic%d></A>\n<HR><H3>%s</H3>\n\n", curCat++,
            yytext+4);
  return TOK_HEADING;
}
^"%"{4}"%"+[^%]+"%"+
[^%\n \t,"{}\\=~]+ {  /*"*/
/*[^%\n \t,"{}\\][^"=,\\{}]* { */
/* (\"|\{)[^"\\{}]* {  */
 /* count the newlines */
 char* ptr = yytext; while (*ptr) { if (*ptr == '\n') linenumber++; ptr++; }
 if (yydebug) printf("\n B: Got <%s>\n", yytext);
 yylval = strdup(yytext);
 return TOK_TEXT;
} 

%%
char buf[BUFSIZE];
char name[BUFSIZE], pagetitle[BUFSIZE], email[BUFSIZE], homepage[BUFSIZE];
char** indices, **strings;
int nStrings = 0, i;

void store_strings();
void create_toc();
void process_data();
void usage();

int main(int argc, char* argv[]) {
  char* ptr;
  i = 1;
  if (argc < 2) usage();
  /* support for more arguments later */
  /* remember to do i++ if an option is actually processed */
  while (argv[i][0] == '-') {
    switch (argv[i][1]) {
      case 'v':
        printf("  Bib2html version 1.2.1.\n"); exit(0); break;
      case 'h':
      case '-':  
      default:
        usage(); exit(0); break;
     }
  }
        
  /* open the input file */
  yyin = fopen(argv[1], "r");
  if (!yyin) { perror("Failed to open input file"); exit(1); }
  strcpy(buf, argv[1]);
  ptr = strstr(buf, ".bib");
  if (ptr) *ptr = 0; else usage();
  strcat(buf, ".html");
  /* set output file */
  fout = fopen(buf, "w");
  if (!fout) { perror("Failed to open output file"); exit(1); }

  /* Set defaults */
  /* Patch contributed by hanwen@cs.uu.nl:
   * getlogin() may return null. */
  if (getlogin()) strcpy(name, getlogin());
  else strcpy(name, "(Name unknown)");
  strcpy(pagetitle, "Bibliography");
  strcpy(homepage, "http://www.litech.org/~wkiri/bib2html/");
  strcpy(email, name);
  strcat(email, "@");
  if (!gethostname(buf, BUFSIZE)) {
    strcat(email, buf);
    if (!getdomainname(buf, BUFSIZE)) {
      strcat(email, ".");
      strcat(email, buf);
    }
  }

  /**** Output the header ****/
  while (fscanf(yyin, "%% %s", buf)) {
    if (!strcasecmp(buf, "Name:"))
      { fgets(name, BUFSIZE, yyin); name[strlen(name)-1] = 0; }
    else if (!strcasecmp(buf, "Title:"))
      { fgets(pagetitle, BUFSIZE, yyin); pagetitle[strlen(pagetitle)-1] = 0; }
    else if (!strcasecmp(buf, "URL:"))
      { fgets(homepage, BUFSIZE, yyin); homepage[strlen(homepage)-1] = 0; }
    else if (!strcasecmp(buf, "Email:")) {
      fgets(email, BUFSIZE, yyin);
      /* bibtex doesn't allow the @, even in a comment! */
      ptr = strstr(email, "%");
      *ptr = '@';
      email[strlen(email)-1] = 0;
      break;
    } else if (!strcmp(buf, "-*-")) 
       { fgets(buf, BUFSIZE, yyin); }
    else if (!strcmp(buf, "%%%")) /* section heading, end now */
       break;
    else /* unrecognized, print an error message? */
       { printf("Unrecognized: <%s>\n", buf); break; }
  }
  
  fprintf(fout, "<HTML><HEAD>\n<TITLE>%s's %s</TITLE>\n</HEAD>\n\n"
	  "<BODY BGCOLOR=#FFFFFF>\n<CENTER><FONT SIZE=+1>\n"
	  "%s : %s<BR></FONT>\n"
	  "<I>Automatically generated from a bibfile<BR>\n"
	  "If you notice any errors or broken links, please let me know:<BR>"
	  "<A HREF=\"mailto:%s\">%s</A></I><P>\n"
	  "</CENTER><P>\n\n<HR>\n", name, pagetitle, name, pagetitle, email, email);
  store_strings();
  create_toc();
  process_data();

  /**** Output the footer ****/
  fprintf(fout, "<HR>\n<ADDRESS><CENTER><A HREF=\"%s\">%s</A>"
	  "&lt; Email: <A HREF=\"mailto:%s\">%s</A> &gt;</ADDRESS>\n"
	  "</BODY>\n</HTML>\n", homepage, name, email, email);
  fclose(yyin);
  fclose(fout);
  return 0;
}

/* Store all the strings in a pseudo-associative array */
void store_strings()
{
  /* First, count them */
  while (fgets(buf, BUFSIZE, yyin))
    if (strstr(buf, "@string") || strstr(buf, "@String")) nStrings++;
  printf("Read and stored %d @strings.\n", nStrings);
  printf("Any output after this point consists of items not recognized by the parser.\n Please send such output to wkiri@litech.org along with the relevant\n section of the bibfile that caused it so that I can update this program.\n------\n");
  rewind(yyin);
  indices = (char**)malloc(nStrings*sizeof(char*)); assert(indices);
  strings = (char**)malloc(nStrings*sizeof(char*)); assert(strings);
  for (i=0; i<nStrings; i++) {
    indices[i] = (char*)malloc(BUFSIZE); assert(indices[i]);
    strings[i] = (char*)malloc(BUFSIZE); assert(strings[i]);
  }
  i = 0;
  while (fgets(buf, BUFSIZE, yyin))
    if (strstr(buf, "@string") || strstr(buf, "@String")) {
      int j = 0, ind = 0, str = 0;

      if (i > nStrings) /* error! */
	{ fprintf(stderr, "Too many strings?"); exit(1); }
      
      /* Go up to and consume the { */
      while (buf[j] && buf[j] != '{') j++; j++;
      /* Pick up whatever's there until a space or = */
      while (buf[j] != ' ' && buf[j] != '=')
	{ indices[i][ind++] = buf[j]; j++; }
      indices[i][ind] = 0; /* null-terminate */
      /* Consume up to and including the " or { */
      /* Better not have mixed types going on! */
      while (buf[j] != '"' && buf[j] != '{') j++; j++;
      /* Pick up the contents of the string */
      while (buf[j] != '"' && buf[j] != '}')
	{ strings[i][str++] = buf[j]; j++; }
      strings[i][str] = 0; /* null-terminate */
      /*  strtok(buf, "{");
	  strcpy(indices[i], strtok(NULL, " "));
	  strtok(NULL, "\""); strcpy(strings[i], strtok(NULL, "\""));
	  i++;
      */
/*     fprintf(stderr, "Stored index <%s>, string <%s>\n", indices[i], strings[i]);*/
      i++; /* next entry */
    }
  rewind(yyin);
}

/* Craft the table of contents by looking for triple %%% in the bibfile */
void create_toc()
{
  i = 0;
  fprintf(fout, "<!-- Table of contents -->\n<CENTER>\n"
	  "<TABLE CELLPADDING=5><TR><TD ALIGN=LEFT><FONT SIZE=+2><B>Topics:</B></FONT></TD>\n"
	  "<TD><TABLE CELLPADDING=10>\n");
  while (fscanf(yyin, "%s", buf) != EOF) {
    if (!strcmp(buf, "%%%")) {
      if (i%3 == 0) fprintf(fout, "<TR>");
      fprintf(fout, "<TD><A HREF=\"#topic%d\">", i++);
      fscanf(yyin, "%s", buf);
      do {
	fprintf(fout, " %s", buf);
      } while (fscanf(yyin, "%s", buf) && strcmp(buf, "%%%"));
      fprintf(fout, "</A></TD>\n");
      if (i%3 == 0) fprintf(fout, "</TR>\n");
    }
    else fgets(buf, BUFSIZE, yyin);
  }
  if (i%3 == 1) fprintf(fout, "<TD></TD><TD></TD></TR>\n");
  else if (i%3 == 2) fprintf(fout, "<TD></TD></TR>\n");
  fprintf(fout, "</TABLE></TD></TR></TABLE></CENTER><P><HR>\n");
  rewind(yyin);
}	      
	      
void process_data()
{
  /*  yylex();*/
  yydebug = 0;
  yyparse();
}

/* Remove ~ from author.
 * Remove {} from title.
 * Turn '--' into '-'.
 * Handle special characters.
 */
void clean(char* str)
{
  char* sptr = str;
  int i;
  while (*sptr) {
    switch (*sptr) {
    case '~':  *sptr = ' '; break;
    case '{':
    case '}':
      i = 0; while (sptr[i]) { sptr[i] = sptr[i+1]; i++; }
      break;
      /* Don't strip $; leave math unmangled. */
      /*    case '$':
      i = 0;
      while (sptr[i] && sptr[i+1] != '$') { sptr[i] = sptr[i+1]; i++; }
      if (sptr[i+1] == '$')
	while (sptr[i+1]) { sptr[i] = sptr[i+2]; i++; }
	break;*/
    case '-':
      if (sptr[1] == '-') 
	{ i = 0; while (sptr[i]) { sptr[i] = sptr[i+1]; i++; } }
      break;
    case '\\':
      switch (sptr[1]) {
      case '&':
	i = 0; while (sptr[i]) { sptr[i] = sptr[i+1]; i++; } break;
      case '-':
	i = 0; while (sptr[i]) { sptr[i] = sptr[i+2]; i++; } break;
      case '"':
	i = 0; while (sptr[i]) { sptr[i] = sptr[i+1]; i++; } break;
      default:
	*sptr = sptr[3];
	i = 1; while (sptr[i]) { sptr[i] = sptr[i+4]; i++; }
	break;
      }
      break;
    }
    sptr++;
  }
}

char* find_string(char* str)
{
  char* retval = str;
  /*  printf("looking up <%s>\n", str);*/
  for (i=0; i<nStrings; i++)
    if (!strcasecmp(str, indices[i])) { retval = strdup(strings[i]); break; }
  /*  printf("returning <%s>\n", retval);*/
  return retval;
}

void usage()
{
  printf("usage: bib2html <bib file>\n");
  exit(1);
}

/*
 * $Log: bib2html.l,v $
 * Revision 1.11  2003/11/30 22:36:03  wkiri
 * Handles custom fields (unrecognized) by printing a message.
 * Allows () instead of {} to delineate @entries.
 * Permits text outside @entries.
 *
 * Revision 1.10  2003/11/11 04:16:58  wkiri
 * Major major changes.  Now using bison.
 *
 * Revision 1.9  2002/07/09 02:30:28  wkiri
 * Header lines can be terminated with whitespace.
 *
 * Revision 1.8  2002/07/07 18:22:24  wkiri
 * Allow @string contents to be enclosed in {}.
 *
 * Revision 1.7  2002/07/07 17:20:40  wkiri
 * Handles nested escaped quotes better.
 * Also allows @String (not just @string).  Not sure if this is valid bib.
 * Got rid of strtok nastiness for parsing in @strings.
 *
 * Revision 1.6  2002/07/01 05:08:45  wkiri
 * Added booklet type.
 * Fields: comment is ignored (treated as notes), isbn is ignored.
 * Correctly handles escaped double-quotes inside a quoted string.
 * Better handling of terminating } in weird places.
 * Does not strip tildes out of URLs. (Aie!)
 * Adds final . to author fields without it.
 * More careful checking; can handle missing author, title, or year.
 * Prints 'edition' after the edition.
 *
 * Revision 1.5  2002/07/01 03:10:49  wkiri
 * Added MANUAL and ORGANIZATION and updated the nEntryTypes.
 * Should handle nested {}'s.
 *
 * Revision 1.4  2002/01/30 15:44:30  wkiri
 * Was missing the final entry; fixed.
 * Was requiring that closing } be the first char on the line; fixed.
 *
 * Revision 1.3  2001/11/09 15:05:26  wkiri
 * Added INBOOK type.
 *
 * Revision 1.2  2001/01/17 05:18:28  wkiri
 * Gracefully handles lack of header info, and guesses values when needed.
 *
 * Revision 1.1  2000/11/10 23:19:10  wkiri
 * Initial revision
 *
 * Revision 1.2  2000/11/02 03:49:50  wkiri
 * Final version, it works!
 *
 * Revision 1.1  2000/11/01 23:26:46  wkiri
 * Initial revision
 *
 */
