%{
  #define YYSTYPE char*
  #define YYDEBUG 1
  #define BUFSIZE 512
  #include <stdio.h>
  #include <string.h>
  #include <assert.h>
  #include "bibentry.h"

  extern FILE* fout;
  extern int linenumber;
  extern char* find_string(char* str);
  void yyerror(const char *str) {
   fprintf(stderr,"Error on line %d: %s\n"
	   "Please ensure that bibtex parses your .bib file cleanly\n"
	   "  (it may provide a more informative syntax error).\n"
	   "If you find that bibtex parses your file but bib2html does not,\n"
	   "  please send the problematic entry to wkiri@litech.org.\n", 
	   linenumber, str);
  }
  char* assemble(char* s1, char* s2, char* s3);
  void store_field(char* fieldtype, char* contents);
  void output_entry();
  ENTRY entry;
  char* curfieldtype;
%}

%token  TOK_TYPE TOK_FIELDTYPE TOK_TEXT TOK_WS TOK_COMMENT
/* Entry fields */
%token  TOK_AUTHOR TOK_ADDRESS TOK_BOOKTITLE TOK_CHAPTER TOK_COMMENT
%token  TOK_EDITION TOK_EDITOR TOK_HOWPUB TOK_INSTITUTION TOK_ISBN
%token  TOK_JOURNAL TOK_KEY TOK_MONTH TOK_NOTE TOK_NOTES 
%token  TOK_NUMBER TOK_ORGANIZATION TOK_PAGES TOK_PDF TOK_PS 
%token  TOK_PUBLISHER TOK_SCHOOL TOK_SERIES TOK_TITLE TOK_TYPE TOK_URL 
%token  TOK_VOLUME TOK_YEAR 
/* Punctuation */
%token  TOK_OPENCURLY TOK_CLOSECURLY TOK_OPAREN TOK_CPAREN
%token  TOK_QUOTE TOK_SLASH TOK_COMMA TOK_EQ
%token  TOK_TILDE 
/* Section headers */
%token  TOK_HEADING
%%

bibfile: entries maybewhitespace
;

whitespace: TOK_WS
| whitespace TOK_WS
;

maybewhitespace: /* empty */
| whitespace 
;

entries: /* empty */
| entries maybewhitespace TOK_TEXT /* permit unparsed text outside
				      entries */
| entries maybewhitespace entry
| entries maybewhitespace comments
| entries maybewhitespace heading
;

/* headings must be terminated by a newline */
heading: TOK_HEADING { fprintf(fout, "%s", $1); } 
;

/* comments are ignored */
comments: TOK_COMMENT
| comments TOK_COMMENT
;

entry: TOK_OPENCURLY { entry.type = strdup($1); } 
  TOK_COMMA contents TOK_CLOSECURLY { output_entry(); }
| TOK_OPAREN { entry.type = strdup($1); } 
  TOK_COMMA contents TOK_CPAREN { output_entry(); }
;

contents: field
| contents TOK_COMMA field
;

field: maybewhitespace /* essentially, empty */
| maybewhitespace TOK_TEXT { curfieldtype = $2; }
   maybewhitespace TOK_EQ maybewhitespace 
  fieldcontents maybewhitespace { store_field($2, $7); }
;

/* how to handle this? {"a} */
fieldcontents: TOK_QUOTE fieldtext TOK_QUOTE { $$ = $2; }
| TOK_OPENCURLY fieldtext TOK_CLOSECURLY { $$ = $2; }
| TOK_OPENCURLY curlytext TOK_CLOSECURLY { $$ = $2; }
| TOK_TEXT { $$ = find_string($1); }
;

fieldtext: /* empty */ { $$ = 0; }
| fieldtext whitespace  { 
  $$ = assemble($1, $2, 0);  if ($1) free($1);  if ($2) free($2);
  if (yydebug) printf("<<%s>>\n", $$); 
}
| fieldtext TOK_TEXT  { 
  $$ = assemble($1, $2, 0);  if ($1) free($1);  if ($2) free($2);
  if (yydebug) printf("<<%s>>\n", $$); 
}
| fieldtext TOK_COMMA {
  $$ = assemble($1, $2, 0);  if ($1) free($1);
  if (yydebug) printf("<<%s>>\n", $$); 
}
| fieldtext TOK_EQ {
  $$ = assemble($1, $2, 0);  if ($1) free($1);
  if (yydebug) printf("<<%s>>\n", $$); 
}
| fieldtext TOK_TILDE {
  /* tilde is interpreted as a non-breaking space, unless we're in a URL */
  if (strcasecmp(curfieldtype, "url"))
    { $$ = assemble($1, "&nbsp;", 0);  if ($1) free($1); }
  else
    { $$ = assemble($1, "~", 0);  if ($1) free($1); }
  if (yydebug) printf("<<%s>>\n", $$); 
}
| fieldtext TOK_SLASH slashitem {
  /* slash modifies the following letter, so don't pass through */
  $$ = assemble($1, $3, 0);  if ($1) free($1);  if ($3) free($3);
  if (yydebug) printf("<<%s>>\n", $$); 
}
| fieldtext TOK_OPENCURLY fieldtext TOK_CLOSECURLY {
  /* drop curlys */
  $$ = assemble($1, $3, 0);  if ($1) free($1);  if ($3) free($3);
  if (yydebug) printf("<<%s>>\n", $$); 
}
| fieldtext TOK_OPENCURLY curlytext fieldtext TOK_CLOSECURLY {
  /* drop curlys */
  $$ = assemble($1, $3, $4);  if ($1) free($1); free($3); if ($4) free($4);
  if (yydebug) printf("<<%s>>\n", $$); 
}
;

/* curlytext is allows, e.g. " without a \ */
/* Only allow single words in {} if " is used */
curlytext: TOK_QUOTE TOK_TEXT {
  /* e.g. {"a} */
  $$ = assemble($1, $2, 0);  free($1); free($2);
  if (yydebug) printf("<<%s>>\n", $$); 
}
| fieldtext TOK_QUOTE TOK_TEXT {
  /* e.g. {Blah"aha} */
  $$ = assemble($1, $2, $3);  if ($1) free($1); free($2); free($3);
  if (yydebug) printf("<<%s>>\n", $$); 
}
| fieldtext TOK_QUOTE TOK_OPENCURLY TOK_TEXT TOK_CLOSECURLY {
  /* e.g. {Blah"{a}} */
  $$ = assemble($1, $2, $4);  if ($1) free($1); free($2); free($4); 
  if (yydebug) printf("<<%s>>\n", $$); 
}
;

/* for some punctuation, just ignore the \ */
/* todo: for ' and ", should generate accented characters */
/* should also handle, e.g. \'{e} */
slashitem: TOK_TEXT { $$ = $1; if (yydebug) printf("<<%s>>\n", $$); }
/* \\: preserve a single \ */
| TOK_SLASH { $$ = $1; if (yydebug) printf("<<%s>>\n", $$); }
| TOK_WS { $$ = $1; if (yydebug) printf("<<%s>>\n", $$); }
| TOK_TILDE { 
  $$ = $1; if (yydebug) printf("<<%s>>\n", $$); 
}
| TOK_QUOTE { 
  /* pass through unchanged */
  $$ = $1; if (yydebug) printf("<<%s>>\n", $$); 
}
| TOK_QUOTE TOK_TEXT { 
  /* this makes an umlaut {\"a}. for now, just preserve the " */
  $$ = assemble("\"", $2, 0);  if ($1) free($1);
  if (yydebug) printf("<<%s>>\n", $$); 
}
;


%%

char* assemble(char* s1, char* s2, char* s3) {
  int len = 1;
  char* str; 
  if (s1) len += strlen(s1);
  if (s2) len += strlen(s2);
  if (s3) len += strlen(s3);
  str = (char*)malloc(len); assert(str); str[0] = 0;
  if (s1) { strcat(str, s1); }
  if (s2) { strcat(str, s2); }
  if (s3) { strcat(str, s3); }
  if (yydebug) printf("ass: <<%s>>\n", str); 
  return str;
}

/* Build the data structure. */
/* Is bibtex case-sensitive for these fields? */
void store_field(char* fieldtype, char* contents) {
  if (contents == NULL) {
    if (yydebug) printf(".....skipping null <%s>\n", fieldtype);
    return;
  } else if (yydebug)  
    printf(".....storing <%s, %s>\n", fieldtype, contents);

  if (!strcasecmp(fieldtype, "abstract")) {
    entry.fields[ABSTRACT] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "address")) {
    entry.fields[ADDRESS] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "author")) {
    entry.fields[AUTHOR] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "booktitle")) {
    entry.fields[BOOKTITLE] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "chapter")) {
    entry.fields[CHAPTER] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "comment")) {
    /*    entry.fields[COMMENT] = strdup(contents);*/
  } else if (!strcasecmp(fieldtype, "edition")) {
    entry.fields[EDITION] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "editor")) {
    entry.fields[EDITOR] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "howpublished")) {
    entry.fields[HOWPUB] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "institution")) {
    entry.fields[INST] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "isbn")) {
    entry.fields[ISBN] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "journal")) {
    entry.fields[JOURNAL] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "key")) {
    entry.fields[KEY] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "month")) {
    entry.fields[MONTH] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "note")) {
    entry.fields[NOTE] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "notes")) {
    entry.fields[NOTES] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "number")) {
    entry.fields[NUMBER] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "organization")) {
    entry.fields[ORG] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "pages")) {
    entry.fields[PAGES] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "pdf")) {
    entry.fields[PDF] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "ps")) {
    entry.fields[PS] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "publisher")) {
    entry.fields[PUB] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "school")) {
    entry.fields[SCHOOL] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "series")) {
    entry.fields[SERIES] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "title")) {
    entry.fields[TITLE] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "type")) {
    entry.fields[TYPE] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "url")) {
    entry.fields[URL] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "volume")) {
    entry.fields[VOLUME] = strdup(contents);
  } else if (!strcasecmp(fieldtype, "year")) {
    entry.fields[YEAR] = strdup(contents);
  } else if (strstr(fieldtype, "OPT") == fieldtype) {
    /* do nothing */
  } else {
    printf("Note: unrecognized field type: <%s>\n", fieldtype);
  }
  free(contents);
}

/* output the entry in HTML */
void output_entry() {
  int i;
  if (yydebug) printf("ENTRY (%s): \n", entry.type);
  if (!strcasecmp(entry.type, "misc")) return;
  /* manuals don't have authors */
  if (!strcasecmp(entry.type, "manual")) 
    entry.fields[AUTHOR] = strdup(entry.fields[KEY]);

  /* They all should have author, title, and year. */
  /* But let's check, just in case. */
  if (entry.fields[AUTHOR]) {
    if (entry.fields[AUTHOR][strlen(entry.fields[AUTHOR])-1] == '.')
      fprintf(fout, "%s\n", entry.fields[AUTHOR]);
    else
      fprintf(fout, "%s.\n", entry.fields[AUTHOR]);
  }
  else if (entry.fields[EDITOR]) {
    if (entry.fields[EDITOR][strlen(entry.fields[EDITOR])-1] == '.')
      fprintf(fout, "%s\n", entry.fields[EDITOR]);
    else
      fprintf(fout, "%s.\n", entry.fields[EDITOR]);
  }
  if (entry.fields[TITLE]) {
    if (entry.fields[URL] && strcasecmp(entry.type, "inbook"))
      fprintf(fout, "<B><A HREF=\"%s\">%s</A>.</B>\n",
	      entry.fields[URL], entry.fields[TITLE]);
    else if (entry.fields[URL] && entry.fields[CHAPTER])
      fprintf(fout, "<B><A HREF=\"%s\">%s</A>.</B>\n",
	      entry.fields[URL], entry.fields[CHAPTER]);
    else if (strcasecmp(entry.type, "inbook"))
      fprintf(fout, "<B>%s.</B>\n", entry.fields[TITLE]);
  } else if (entry.fields[CHAPTER])
    fprintf(fout, "<B>%s.</B>\n", entry.fields[CHAPTER]);
  if (entry.fields[YEAR]) fprintf(fout, "%s.\n", entry.fields[YEAR]);
    
  /* main entry type switch */
  if (!strcasecmp(entry.type, "book")) {
    if (entry.fields[PUB] && entry.fields[ADDRESS])
      fprintf(fout, "%s, %s.\n", entry.fields[PUB],  entry.fields[ADDRESS]);
    else if (entry.fields[PUB]) 
      fprintf(fout, "%s.\n", entry.fields[PUB]);
    else if (entry.fields[ADDRESS]) 
      fprintf(fout, "%s.\n", entry.fields[ADDRESS]);
    if (entry.fields[EDITION]) 
      fprintf(fout, "%s edition.\n", entry.fields[EDITION]);
  } else if (!strcasecmp(entry.type, "inbook")) {
    fprintf(fout, "Chapter in <I>%s.</I>\n", entry.fields[TITLE]);
    if (entry.fields[PUB] && entry.fields[ADDRESS])
      fprintf(fout, "%s, %s.\n", entry.fields[PUB], entry.fields[ADDRESS]);
    else if (entry.fields[PUB]) 
      fprintf(fout, "%s.\n", entry.fields[PUB]);
    else if (entry.fields[ADDRESS]) 
      fprintf(fout, "%s.\n", entry.fields[ADDRESS]);
  } else if (!strcasecmp(entry.type, "article")) {
    /* Contributed by Retvari Gabor, retvari@ttt-atm.ttt.bme.hu */
    if (!entry.fields[PAGES] && !entry.fields[VOLUME])
      fprintf(fout, "<I>%s.</I>\n", entry.fields[JOURNAL]);
    else if (entry.fields[PAGES] && !entry.fields[VOLUME])
      fprintf(fout, "<I>%s,</I> pages %s.\n",
	      entry.fields[JOURNAL], entry.fields[PAGES]);
    else if (!entry.fields[PAGES] && entry.fields[VOLUME])
      fprintf(fout, "<I>%s,</I> vol %s.\n",
	      entry.fields[JOURNAL], entry.fields[VOLUME]);
    else if (entry.fields[NUMBER])
       fprintf(fout, "<I>%s,</I> %s(%s):%s.\n",
	       entry.fields[JOURNAL], entry.fields[VOLUME], entry.fields[NUMBER], entry.fields[PAGES]);
    else 
       fprintf(fout, "<I>%s,</I> %s:%s.\n",
	       entry.fields[JOURNAL], entry.fields[VOLUME], entry.fields[PAGES]);
  } else if (!strcasecmp(entry.type, "inproceedings") ||
	     !strcasecmp(entry.type, "incollection")) {
    if (entry.fields[PAGES])
      fprintf(fout, "in <I>%s,</I> pages %s.\n",
	      entry.fields[BOOKTITLE], entry.fields[PAGES]);
    else
      fprintf(fout, "in <I>%s</I>.\n", entry.fields[BOOKTITLE]);
    if (entry.fields[ORG]) fprintf(fout, "%s.\n", entry.fields[ORG]);
    if (entry.fields[PUB] && entry.fields[ADDRESS])
      fprintf(fout, "%s, %s.\n", entry.fields[PUB], entry.fields[ADDRESS]);
    else if (entry.fields[PUB]) fprintf(fout, "%s.\n", entry.fields[PUB]);
    else if (entry.fields[ADDRESS]) fprintf(fout, "%s.\n", entry.fields[ADDRESS]);
  } else if (!strcasecmp(entry.type, "techreport")) {
    if (entry.fields[TYPE]) fprintf(fout, "%s, ", entry.fields[TYPE]);
    if (entry.fields[NUMBER])  fprintf(fout, "%s, ", entry.fields[NUMBER]);
    if (entry.fields[ADDRESS] && entry.fields[INST])
      fprintf(fout, "%s, %s.\n", entry.fields[INST], entry.fields[ADDRESS]);
    else if (entry.fields[INST])
      fprintf(fout, "%s.\n", entry.fields[INST]);
  } else if (!strcasecmp(entry.type, "phdthesis")) {
    if (entry.fields[SCHOOL] && entry.fields[ADDRESS])
      fprintf(fout, "Ph. D. thesis, %s, %s.\n", 
	      entry.fields[SCHOOL], entry.fields[ADDRESS]);
    else if (entry.fields[SCHOOL])
      fprintf(fout, "Ph. D. thesis, %s.\n", entry.fields[SCHOOL]);
    else fprintf(fout, "Ph. D. thesis.\n");
  } else if (!strcasecmp(entry.type, "mastersthesis")) {
    if (entry.fields[SCHOOL] && entry.fields[ADDRESS])
      fprintf(fout, "Master's thesis, %s, %s.\n", 
	      entry.fields[SCHOOL], entry.fields[ADDRESS]);
    else if (entry.fields[SCHOOL])
      fprintf(fout, "Master's thesis, %s.\n", entry.fields[SCHOOL]);
    else fprintf(fout, "Master's thesis.\n");
  } else if (!strcasecmp(entry.type, "unpublished")) {
    fprintf(fout, "Unpublished.\n");
  } else if (!strcasecmp(entry.type, "manual")) {
    if (entry.fields[ORG] && entry.fields[ADDRESS])
      fprintf(fout, "%s, %s.\n", entry.fields[ORG], entry.fields[ADDRESS]);
    else if (entry.fields[ORG])
      fprintf(fout, "%s.\n", entry.fields[ORG]);
    if (entry.fields[EDITION])
      fprintf(fout, "%s.\n", entry.fields[EDITION]);
  } else if (!strcasecmp(entry.type, "booklet")) {
    if (entry.fields[HOWPUB]) fprintf(fout, "%s.\n", entry.fields[HOWPUB]);
  } else {
    printf("Unknown entry type <%s>.\n", entry.type);
  }
  if (entry.fields[NOTE]) fprintf(fout, "(%s)\n", entry.fields[NOTE]);
  /* Contributed by David Deharbe, david.deharbe@loria.fr */
  if (entry.fields[PS] || entry.fields[PDF]) {
    fprintf(fout, "[ ");
    if (entry.fields[PS] ) fprintf(fout, "<a href=\042%s\042>.ps</a>", entry.fields[PS]);
    if (entry.fields[PS] && entry.fields[PDF]) fprintf(fout, " | ");
    if (entry.fields[PDF] ) fprintf(fout, "<a href=\042%s\042>.pdf</a>", entry.fields[PDF]);
    fprintf(fout, " ]");
  }
  if (entry.fields[ABSTRACT]) fprintf(fout, "<p>\n<blockquote><font size=-1>\n%s\n</font></blockquote>", entry.fields[ABSTRACT]);
  fprintf(fout, "<P>\n\n");
  /* Then wipe it (free all malloc'd fields) */
  for (i=0; i<nFields; i++) if (entry.fields[i]) { 
    free(entry.fields[i]); entry.fields[i] = 0; 
  }
  free(entry.type);
}
